/*
Watcom C/C++ Automated MAKE Utility v1.02.
Written by Tom Torfs. Donated to the public domain.
Last update: July 17th, 1997.
Last changes: * support for Windows NT
              * works properly with subdirectories now
*/

#ifndef __WATCOMC__
#error You need Watcom C to compile WAMAKE.C.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>

enum {DOS,DOS4G,OS2,OS2V2,NT,NUMTARGETS};

const char *targetname[NUMTARGETS]
   = {"DOS","DOS4G","OS2","OS2V2","NT"};

const char *targetdefine[NUMTARGETS]
   = {"","__DOS__","","__OS2__","__NT__"};
   
const char *wasm[NUMTARGETS]
   = {"WASM","WASM","WASM","WASM","WASM"};
const char *wcc[NUMTARGETS]
   = {"WCC","WCC386","WCC","WCC386","WCC386"};
const char *wpp[NUMTARGETS]
   = {"WPP","WPP386","WPP","WPP386","WPP386"};
const char *wlink[NUMTARGETS]
   = {"WLINK","WLINK","WLINK","WLINK","WLINK"};
   
const char linkfile[] = "~WAMAKE~.LNK";
   
static char configfile[_MAX_PATH] = "WAMAKE.WAM";
static char drive[_MAX_DRIVE],dir[_MAX_DIR],fname[_MAX_FNAME],ext[_MAX_EXT];
static char wccopt[128],wasmopt[128];
static char buf[256];
static char exename[_MAX_PATH]="",srcname[_MAX_PATH],objname[_MAX_PATH];

int main(int argc, char *argv[])
{
   FILE *fp,*linkf;
   struct find_t ffblk;
   unsigned short srcdate,srctime,objdate,objtime;
   int len;
   int relink = 0, nolink = 0, forced;
   int target
#if defined(__OS2__)
       = OS2;
#elif defined(__NT__)
       = NT;
#else
       = DOS;
#endif
   
   puts("Watcom C/C++ Automated MAKE Utility v1.02 - by Tom Torfs\n");
   
   if (argc>2 || (argc==2 && strchr(argv[1],'?')!=NULL))
   {
      puts("Usage: WAMAKE <configfile>[.WAM]\n\n"
           "Read WAMAKE.DOC for more information.");
      return EXIT_FAILURE;
   }
   
   if (argc>1)
   {
      _fullpath(configfile,argv[1],_MAX_PATH);
      strupr(configfile);
      _splitpath(configfile,drive,dir,fname,ext);
      if (ext[0]=='\0')
         _makepath(configfile,drive,dir,fname,".WAM");
   }
   
   if ((fp=fopen(configfile,"r"))==NULL)
   {
      printf("Error: can't read %s\n",configfile);
      return EXIT_FAILURE;
   }
   
   if ((linkf=fopen(linkfile,"w"))==NULL)
   {
      printf("Error: can't create temporary file %s\n",linkfile);
      fclose(fp);
      return EXIT_FAILURE;
   }
   
   while (1)
   {
      fgets(buf,256,fp);
      if (feof(fp))
         break;
      len = strlen(buf);
      while (len>0 && isspace(buf[len-1]))
         buf[--len] = '\0';
      while (isspace(buf[0]))
         memmove(buf,buf+1,len--);
      if (len==0 || buf[0]==';')
         continue;
      if (buf[0]=='.' && isalpha(buf[1]))
      {
         if (memicmp(buf,".target ",8)==0)
         {
            for (target=0; target<NUMTARGETS; target++)
            if (stricmp(buf+8,targetname[target])==0)
               break;
            if (target==NUMTARGETS)
            {
               printf("Error: invalid target system: %s\n",buf+8);
               fclose(linkf);
               fclose(fp);
               remove(linkfile);
               return EXIT_FAILURE;
            }
            continue;
         }
         if (memicmp(buf,".exename ",9)==0)
         {
            strncpy(exename,buf+9,_MAX_PATH);
            exename[_MAX_PATH-1] = '\0';
            strupr(exename);
            _splitpath(exename,drive,dir,fname,ext);
            if (ext[0]=='\0')
               _makepath(exename,drive,dir,fname,".EXE");
            continue;
         }
         if (memicmp(buf,".wccopt ",8)==0)
         {
            strncpy(wccopt,buf+8,128);
            wccopt[127] = '\0';
            continue;
         }
         if (memicmp(buf,".wasmopt ",8)==0)
         {
            strncpy(wasmopt,buf+8,128);
            wasmopt[127] = '\0';
            continue;
         }
         if (memicmp(buf,".directive ",11)==0)
         {
            fprintf(linkf,"%s\n",buf+11);
            continue;
         }
         if (stricmp(buf,".nolink")==0)
         {
            nolink = 1;
            continue;
         }
         printf("Invalid directive: %s\n",buf);
         fclose(linkf);
         fclose(fp);
         remove(linkfile);
         return EXIT_FAILURE;
      }
      if (buf[0]=='*')
      {
         forced = 1;
         strncpy(srcname,buf+1,_MAX_PATH);
      }
      else
      {
         forced = 0;
         strncpy(srcname,buf,_MAX_PATH);
      }
      srcname[_MAX_PATH-1] = '\0';
      strupr(srcname);
      _fullpath(objname,srcname,_MAX_PATH);
      _splitpath(objname,drive,dir,fname,ext);
      _makepath(objname,"","",fname,".OBJ");
      if (exename[0]=='\0')
         _makepath(exename,"","",fname,".EXE");
      if (stricmp(ext,".LIB")==0)
      {
         fprintf(linkf,"library %s\n",srcname);
         continue;
      }
      if (stricmp(ext,".OBJ")==0)
      {
         fprintf(linkf,"file %s\n",srcname);
         continue;
      }
      fprintf(linkf,"file %s\n",objname);
      if (stricmp(ext,".C")==0)
      {
         strcpy(buf,wcc[target]);
         if (wccopt[0]!='\0')
         {
            strcat(buf," ");
            strcat(buf,wccopt);
         }
      }
      else if (stricmp(ext,".CPP")==0 || stricmp(ext,".CC")==0 || stricmp(ext,".CXX")==0)
      {
         strcpy(buf,wpp[target]);
         if (wccopt[0]!='\0')
         {
            strcat(buf," ");
            strcat(buf,wccopt);
         }
      }
      else if (stricmp(ext,".ASM")==0)
      {
         strcpy(buf,wasm[target]);
         if (wasmopt[0]!='\0')
         {
            strcat(buf," ");
            strcat(buf,wasmopt);
         }
      }
      else
      {
         printf("Invalid module: %s (must be .C/.CC/.CPP/.CXX/.ASM/.OBJ/.LIB)\n",srcname);
         fclose(linkf);
         fclose(fp);
         remove(linkfile);
         return EXIT_FAILURE;
      }
      strcat(buf," /bt=");
      strcat(buf,targetname[target]);
      if (targetdefine[target][0]!='\0')
      {
         strcat(buf," /d");
         strcat(buf,targetdefine[target]);
      }
      strcat(buf," ");
      strcat(buf,srcname);
      if (_dos_findfirst(srcname,_A_NORMAL,&ffblk)!=0)
      {
         _dos_findclose(&ffblk);
         printf("Error: %s not found\n",srcname);
         fclose(linkf);
         fclose(fp);
         remove(linkfile);
         return EXIT_FAILURE;
      }
      else
      {
         srcdate = ffblk.wr_date;
         srctime = ffblk.wr_time;
      }
      _dos_findclose(&ffblk);
      if (_dos_findfirst(objname,_A_NORMAL,&ffblk)!=0)
      {
         objdate = 0;
         objtime = 0;
      }
      else
      {
         objdate = ffblk.wr_date;
         objtime = ffblk.wr_time;
      }
      _dos_findclose(&ffblk);
      if (forced || srcdate>objdate || (srcdate==objdate && srctime>objtime))
      {
         remove(objname);
         system(buf);
         if (_dos_findfirst(objname,_A_NORMAL,&ffblk)!=0)
         {
            _dos_findclose(&ffblk);
            puts("\nCompiler/assembler returned bad status.");
            fclose(linkf);
            fclose(fp);
            remove(linkfile);
            return EXIT_FAILURE;
         }
         _dos_findclose(&ffblk);         
         relink = 1;
      }
   }

   /* add default libraries here */

   fprintf(linkf,"name %s\n",exename);
   fprintf(linkf,"system %s\n",targetname[target]);
   
   fclose(linkf);
   fclose(fp);

   if (!nolink)
   {
      if (!relink)
      {
         puts("No modules need recompiling.");
         if (_dos_findfirst(exename,_A_NORMAL,&ffblk)!=0)
         {
            puts("\nRelinking...");
            relink = 1;
         }
         _dos_findclose(&ffblk);
      }

      if (relink)
      {
         remove(exename);
         strcpy(buf,wlink[target]);
         strcat(buf," @");
         strcat(buf,linkfile);
         system(buf);
         if (_dos_findfirst(exename,_A_NORMAL,&ffblk)!=0)
         {
            _dos_findclose(&ffblk);
            puts("\nLinker returned bad status.");
            remove(linkfile);
            return EXIT_FAILURE;
         }
         _dos_findclose(&ffblk);
         printf("\nDone. Created %s (%ld bytes)\n",exename,ffblk.size);
      }
   }
      
   remove(linkfile);
   
   return EXIT_SUCCESS;
}
